import json
import aiohttp
from aiohttp import web
import asyncio
import re
import numpy as np
from typing import List, Dict, Optional, Callable, Coroutine
from joule.utilities import interval_tools, timestamp_to_datetime
from joule.models import DataStream, pipes
from joule.models.data_store.data_store import DataStore, StreamInfo, DbInfo, Interval
from joule.models.data_store import errors
from joule.models.data_store.nilmdb_inserter import Inserter
from joule.models.data_store.nilmdb_helpers import (compute_path,
                                                    check_for_error,
                                                    ERRORS)

Loop = asyncio.AbstractEventLoop


class PathNotFound(Exception):
    pass


class NilmdbStore(DataStore):

    def __init__(self, server: str, insert_period: float, cleanup_period: float):
        self.server = server
        self.decimation_factor = 4
        self.insert_period = insert_period
        self.cleanup_period = cleanup_period
        self.connector = None

    async def initialize(self, streams: List[DataStream]) -> None:
        self.connector = aiohttp.TCPConnector()
        url = "{server}/stream/create".format(server=self.server)
        try:
            async with self._get_client() as session:
                for stream in streams:
                    data = {"path": compute_path(stream),
                            "layout": stream.layout}
                    async with session.post(url, data=data) as resp:
                        await check_for_error(resp, ignore=[ERRORS.STREAM_ALREADY_EXISTS])
        except aiohttp.ClientError:
            raise errors.DataError("cannot contact NilmDB at [%s]" % self.server)

    async def insert(self, stream: DataStream, start: int, end: int, data: np.array) -> None:
        """insert stream data"""
        url = "{server}/stream/insert".format(server=self.server)
        params = {"start": "%d" % start,
                  "end": "%d" % end,
                  "path": compute_path(stream),
                  "binary": '1'}
        async with self._get_client() as session:
            async with session.put(url, params=params,
                                   data=data.tobytes()) as resp:
                if resp.status != 200:
                    if resp.status == 400:
                        error = await resp.json()
                        # nilmdb rejected the data
                        raise errors.DataError(error["message"])
                    raise errors.DataError(await resp.text())  # pragma: no cover

    async def spawn_inserter(self, stream: DataStream,
                             pipe: pipes.Pipe, insert_period=None, retry_interval=0.5) -> asyncio.Task:
        if insert_period is None:
            insert_period = self.insert_period
        inserter = Inserter(self.server, stream,
                            insert_period, self.cleanup_period, self._get_client,
                            retry_interval=retry_interval)
        t = asyncio.create_task(inserter.run(pipe))
        t.set_name("NilmDB Inserter for [%s]" % stream.name)
        return t

    async def extract(self, stream: DataStream, start: Optional[int], end: Optional[int],
                      callback: Callable[[np.ndarray, str, int], Coroutine], max_rows: int = None,
                      decimation_level=None):
        # figure out appropriate decimation level
        if decimation_level is None:
            if max_rows is None:
                decimation_level = 1
            else:
                # find out how much data this represents
                count = await self._count_by_path(compute_path(stream),
                                                  start, end)
                if count > 0:
                    desired_decimation = np.ceil(count / max_rows)
                    decimation_level = 4 ** np.ceil(np.log(desired_decimation) /
                                                    np.log(self.decimation_factor))
                else:
                    # create an empty array with the right data type
                    data = np.array([], dtype=pipes.compute_dtype(stream.layout))
                    await callback(data, stream.layout, 1)
                    return
                # make sure the target decimation level exists and has data
                try:
                    path = compute_path(stream, decimation_level)
                    if (await self._count_by_path(path, start, end)) == 0:
                        # no data in the decimated path
                        raise errors.InsufficientDecimationError("required level is empty")
                except errors.DataError as e:
                    if ERRORS.NO_SUCH_STREAM.value in str(e):
                        # no decimated data or required level does not exist
                        raise errors.InsufficientDecimationError("required level %d does not exist" % decimation_level)
                    # some other error, propogate it up
                    raise e  # pragma: no cover
        elif max_rows is not None:
            # two constraints, make sure we aren't going to return too much data
            count = await self._count_by_path(compute_path(stream, decimation_level),
                                              start, end)
            if count > max_rows:
                raise errors.InsufficientDecimationError("actual_rows(%d) > max_rows(%d)" % (count, max_rows))

        # retrieve data from stream
        path = compute_path(stream, decimation_level)
        if decimation_level > 1:
            layout = stream.decimated_layout
        else:
            layout = stream.layout
        try:
            await self._extract_by_path(path, start, end, layout, callback)
        except aiohttp.ClientError as e:
            raise errors.DataError(str(e))

    async def intervals(self, stream: DataStream, start: Optional[int], end: Optional[int]):
        try:
            return await self._intervals_by_path(compute_path(stream),
                                                 start, end)
        except errors.DataError as e:
            # if the stream hasn't been written to it won't exist in the database
            if ERRORS.NO_SUCH_STREAM.value in str(e):
                return []
            else:
                raise e

    async def consolidate(self, stream: 'DataStream', start: int, end: int, max_gap: int) -> int:
        # remove interval gaps less than or equal to max_gap duration (in us)
        intervals = await self.intervals(stream, start, end)
        if len(intervals) == 0:
            return  # no data, nothing to do
        duration = [intervals[0][0], intervals[-1][1]]

        gaps = interval_tools.interval_difference([duration], intervals)

        if len(gaps) == 0:
            return  # no interval breaks, nothing to do
        small_gaps = [gap for gap in gaps if (gap[1] - gap[0]) <= max_gap]
        # spawn an inserter to close each gap

        info = await self._path_info([stream])
        all_paths = info.keys()
        path = compute_path(stream)
        insert_url = "{server}/stream/insert".format(server=self.server)
        for gap in small_gaps:
            async with self._get_client() as session:
                params = {"start": "%d" % gap[0],
                          "end": "%d" % gap[1],
                          "path": path,
                          "binary": '1'}
                async with session.put(insert_url, params=params,
                                       data=None) as resp:
                    if resp.status != 200:  # pragma: no cover
                        error = await resp.text()
                        raise errors.DataError("NilmDB(d) error: %s" % error)
        return len(small_gaps)

    # TODO: remove path lookup, iterate until the stream decimation isn't found
    async def remove(self, stream, start: Optional[int] = None,
                     end: Optional[int] = None):
        """ remove [start,end] in path and all decimations """
        info = await self._path_info([stream])
        all_paths = info.keys()
        base_path = compute_path(stream)
        regex = re.compile(r"%s~decim-(\d)+$" % base_path)
        decim_paths = list(filter(regex.match, all_paths))
        for path in [base_path, *decim_paths]:
            await self._remove_by_path(path, start, end)

    async def info(self, streams: List[DataStream]) -> Dict[int, StreamInfo]:
        info_dict = await self._path_info(streams)
        # go through each stream and compute the effective DataStreamInfo
        # object by looking at the raw and decimated paths
        streams_info = {}
        for s in streams:
            base_info_list = [info for (path, info) in info_dict.items() if path == compute_path(s)]
            if len(base_info_list) == 0:
                # this stream has no data records, just make up an empty result
                streams_info[s.id] = StreamInfo(None, None, 0, 0, 0)
                continue
            stream_info: StreamInfo = base_info_list[0]
            if stream_info.rows != 0:
                # stream has data, check for decimations
                regex = re.compile(r"%s~decim-(\d)+$" % compute_path(s))
                decim_info_list = [info for (path, info) in info_dict.items() if regex.match(path) is not None]
                for decim_info in decim_info_list:
                    if decim_info.rows == 0:
                        continue  # no data, don't try to look for start and end values
                    stream_info.start = min((stream_info.start, decim_info.start))
                    stream_info.end = max((stream_info.end, decim_info.end))
                    stream_info.bytes += decim_info.bytes
                    stream_info.total_time = max((stream_info.total_time, decim_info.total_time))
            streams_info[s.id] = stream_info
        return streams_info

    async def dbinfo(self) -> DbInfo:
        url = "{server}/dbinfo".format(server=self.server)
        async with self._get_client() as session:
            async with session.get(url) as resp:
                data = await resp.json()
                return DbInfo(**data)

    async def destroy(self, stream: DataStream):
        await self.remove(stream)
        url = "{server}/stream/destroy".format(server=self.server)
        info = await self._path_info([stream])
        all_paths = info.keys()
        base_path = compute_path(stream)
        regex = re.compile(r'%s~decim-(\d)+$' % base_path)
        decim_paths = list(filter(regex.match, all_paths))
        async with self._get_client() as session:
            for path in [base_path, *decim_paths]:
                async with session.post(url, data={"path": path}) as resp:
                    await check_for_error(resp, ignore=[ERRORS.NO_STREAM_AT_PATH])
        return web.Response(text="ok")

    async def destroy_all(self):
        raise Exception("NilmDB backend does not implement erase_all")

    async def _extract_by_path(self, path: str, start: Optional[int], end: Optional[int],
                               layout: str, callback):
        url = "{server}/stream/extract".format(server=self.server)
        params = {"path": path,
                  "binary": 1}
        decimation_factor = 1
        r = re.search(r'~decim-(\d+)$', path)
        if r is not None:
            decimation_factor = int(r[1])
        async with self._get_client() as session:
            # first determine the intervals, use the base path for this
            if path.find("~decim") == -1:
                base_path = path
            else:
                base_path = path[:path.find("~decim")]
            intervals = await self._intervals_by_path(base_path, start, end)
            i = 0
            num_intervals = len(intervals)
            # now extract each interval
            for interval in intervals:
                params["start"] = interval[0]
                params["end"] = interval[1]
                async with session.get(url, params=params) as resp:
                    await check_for_error(resp)
                    # put data into the queue as it arrives
                    reader = pipes.InputPipe(name="outbound", layout=layout, reader=resp.content)
                    while True:
                        try:
                            data = await reader.read()
                            await callback(data, layout, decimation_factor)
                            reader.consume(len(data))
                        except pipes.EmptyPipe:
                            break
                # insert the interval token to indicate a break
                i += 1
                if i < num_intervals:
                    await callback(pipes.interval_token(layout), layout, decimation_factor)

    async def _intervals_by_path(self, path: str, start: Optional[int],
                                 end: Optional[int]) -> List[Interval]:
        url = "{server}/stream/intervals".format(server=self.server)
        base_path = path
        params = {"path": path}
        if start is not None:
            params["start"] = start
        if end is not None:
            params["end"] = end
        intervals = []
        async with self._get_client() as session:
            async with session.get(url, params=params) as resp:
                await check_for_error(resp)
                text = await resp.text()
                if text == '':
                    return []
                for line in text.strip().split("\n"):
                    intervals.append(json.loads(line))

        return consolidate_intervals(intervals)

    async def _remove_by_path(self, path: str, start: Optional[int], end: Optional[int]):
        """remove data from streams"""
        url = "{server}/stream/remove".format(server=self.server)
        params = {"path": path}
        if start is not None:
            params["start"] = start
        if end is not None:
            params["end"] = end
        async with self._get_client() as session:
            async with session.post(url, params=params) as resp:
                await check_for_error(resp, ignore=[ERRORS.NO_SUCH_STREAM])
                await resp.text()  # wait for operation to complete

    async def _count_by_path(self, path, start: Optional[int], end: Optional[int]) -> int:
        url = "{server}/stream/extract".format(server=self.server)
        params = {"path": path,
                  "count": 1}
        if start is not None:
            params["start"] = start
        if end is not None:
            params["end"] = end
        async with self._get_client() as session:
            async with session.get(url, params=params) as resp:
                await check_for_error(resp)
                return int(await resp.text())

    async def _path_info(self, streams: List[DataStream]) -> Dict[str, StreamInfo]:
        """ set path to None to list all streams """
        url = "{server}/stream/list".format(server=self.server)
        paths = []
        if len(streams) > 6:
            paths = "/joule/%"
        else:
            for stream in streams:
                paths += [f'/joule/{stream.id}', f'/joule/{stream.id}~decim%']
        params = {"extended": 1,
                  "path": paths}
        async with self._get_client() as session:
            async with session.get(url, params=params) as resp:
                body = await resp.text()
                if resp.status != 200:  # pragma: no cover
                    raise errors.DataError(body)
                data = json.loads(body)
                info = {}
                for item in data:
                    info[item[0]] = StreamInfo(start=item[2],
                                               end=item[3],
                                               rows=item[4],
                                               bytes=item[4] * bytes_per_row(item[1]),
                                               total_time=item[5])
                return info

    def _get_client(self):
        return aiohttp.ClientSession(connector=self.connector, connector_owner=False)

    async def close(self):
        if self.connector is not None:
            await self.connector.close()


def bytes_per_row(layout: str) -> int:
    try:
        ltype = layout.split('_')[0]
        lcount = int(layout.split('_')[1])
        if ltype.startswith('int'):
            return (int(ltype[3:]) // 8) * lcount + 8
        elif ltype.startswith('uint'):
            return (int(ltype[4:]) // 8) * lcount + 8
        elif ltype.startswith('float'):
            return (int(ltype[5:]) // 8) * lcount + 8
        else:
            raise ValueError("bad layout %s" % layout)
    except (ValueError, IndexError):
        raise ValueError("bad layout: %s" % layout)


def consolidate_intervals(intervals):
    # consolidate intervals that touch
    consolidated_intervals = []
    if len(intervals) <= 1:
        return intervals
    prev_interval = intervals[0]
    for i in range(1, len(intervals)):
        if intervals[i][0] == prev_interval[1]:
            # extend prev_interval
            prev_interval[1] = intervals[i][1]
        else:
            consolidated_intervals.append(prev_interval)
            prev_interval = intervals[i]
    consolidated_intervals.append(prev_interval)
    return consolidated_intervals



import click
from tabulate import tabulate
import datetime
import asyncio

from joule.models import data_stream
from joule import errors
from joule.cli.config import pass_config


@click.command(name="info")
@click.option('-e', "--elements", help="show element information", is_flag=True)
@click.argument("path")
@pass_config
def cli_info(config, elements, path):
    """Display stream information."""
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            _run(config.node, path, elements))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _run(node, path, show_elements):
    my_stream = await node.data_stream_get(path)
    # display stream information
    click.echo()
    click.echo("Data Stream Information:")
    click.echo("\tName:         %s" % my_stream.name)
    click.echo("\tDescription:  %s" % _optional_field(my_stream.description))
    click.echo("\tDatatype:     %s" % my_stream.datatype.lower())
    click.echo("\tKeep:         %s" % _display_keep(my_stream.keep_us))
    click.echo("\tDecimate:     %s" % _display_decimate(my_stream.decimate))
    click.echo()
    # display information from the data store
    click.echo("\tStatus:       %s" % _display_status(my_stream.locked, my_stream.active))
    my_info = await node.data_stream_info(path)
    click.echo("\tStart:        %s" % _display_time(my_info.start))
    click.echo("\tEnd:          %s" % _display_time(my_info.end))
    click.echo("\tRows:         %d" % my_info.rows)
    click.echo()
    # display element information
    if show_elements:
        elem_data = []
        for element in my_stream.elements:
            elem_data.append([element.name,
                              _optional_field(element.units),
                              element.display_type.lower(),
                              _display_bounds(element.default_max,
                                              element.default_min)
                              ])
        # display element information
        click.echo(tabulate(elem_data, headers=["Name", "Units", "Display", "Min,Max"],
                            stralign="center",
                            tablefmt="fancy_grid"))


def _display_decimate(decimate: bool) -> str:
    if decimate:
        return "yes"
    else:
        return "no"


def _display_keep(keep: int) -> str:
    if keep == data_stream.DataStream.KEEP_NONE:
        return "no data"
    if keep == data_stream.DataStream.KEEP_ALL:
        return "all data"
    return str(datetime.timedelta(microseconds=keep))


def _display_status(locked: bool, active: bool) -> str:
    if active:
        return click.style("\u25CF ", fg="green") + "[active]"
    elif locked:
        return click.style("\u25CF ", fg="cyan") + "[configured]"
    else:
        return "[idle]"


def _display_time(time: int) -> str:
    if time is None:
        return u"\u2014"  # emdash
    return str(datetime.datetime.fromtimestamp(time/1e6))


def _display_bounds(max_: float, min_: float) -> str:
    if max_ is None and min_ is None:
        return "auto"
    elif max_ is None:
        return "%d,[auto]" % min_
    elif min_ is None:
        return "[auto],%d" % max_
    else:
        return "%d,%d" % (min_, max_)


def _optional_field(value: str) -> str:
    if value is None or value == "":
        return u"\u2014"  # emdash
    else:
        return value



class NodeConfig:
    def __init__(self, name, url, key):
        self.name = name
        self.url = url
        self.key = key

    def to_json(self):
        return {
            "name": self.name,
            "url": self.url,
            "key": self.key
        }



import click
import asyncio
import datetime

from joule import errors
from joule.api import BaseNode
from joule.cli.config import Config, pass_config


@click.command(name="info")
@click.argument("name")
@pass_config
def cli_info(config: Config, name: str):
    """Display module information"""
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            _run(config.node, name))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _run(node: BaseNode, name: str):
    module = await node.module_get(name)
    # display module information
    click.echo()
    if module.is_app:
        click.echo("This is module is a Data App\n")
    click.echo("Name:\n\t%s" % module.name)
    if len(module.description) > 0:
        click.echo("Description:\n\t%s" % module.description)
    click.echo("Inputs:")
    if len(module.inputs) == 0:
        click.echo("\t--none--")
    else:
        for (name, loc) in module.inputs.items():
            click.echo("\t%s: %s" % (name, loc))
    click.echo("Outputs:")
    if len(module.outputs) == 0:
        click.echo("\t--none--")
    else:
        for (name, loc) in module.outputs.items():
            click.echo("\t%s:\t%s" % (name, loc))
    click.echo("CPU Usage:\n\t%0.2f%%" % module.statistics.cpu_percent)
    click.echo("Memory Usage:\n\t%0.2f%%" % module.statistics.memory_percent)
    now = datetime.datetime.now().timestamp()
    delta = datetime.timedelta(seconds=now - module.statistics.create_time)

    click.echo("Uptime:\n\t%s" % delta)
    click.echo("")





from aiohttp import web
import json
from sqlalchemy.orm import Session
from joule.models import EventStream, EventStore, event_stream

from joule.models import folder, Folder
from joule.errors import ConfigurationError


async def info(request: web.Request):
    db: Session = request.app["db"]
    event_store: EventStore = request.app["event-store"]
    if 'path' in request.query:
        my_stream = folder.find_stream_by_path(request.query['path'], db, stream_type=EventStream)
    elif 'id' in request.query:
        my_stream = db.query(EventStream).get(request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if my_stream is None:
        return web.Response(text="stream does not exist", status=404)
    stream_info = await event_store.info([my_stream])
    return web.json_response(my_stream.to_json(stream_info))


async def move(request: web.Request):
    db: Session = request.app["db"]
    if request.content_type != 'application/json':
        return web.Response(text='content-type must be application/json', status=400)
    body = await request.json()
    # find the stream
    if 'src_path' in body:
        my_stream = folder.find_stream_by_path(body['src_path'], db, stream_type=EventStream)
    elif 'src_id' in body:
        my_stream = db.query(EventStream).get(body["src_id"])
    else:
        return web.Response(text="specify a source id or a path", status=400)
    if my_stream is None:
        return web.Response(text="stream does not exist", status=404)
    # find or create the destination folder
    if 'dest_path' in body:
        try:
            destination = folder.find(body['dest_path'], db, create=True)
        except ConfigurationError as e:
            return web.Response(text="Destination error: %s" % str(e), status=400)
    elif 'dest_id' in body:
        destination = db.query(Folder).get(body["dest_id"])
    else:
        return web.Response(text="specify a destination", status=400)
    # make sure name is unique in this destination
    existing_names = [s.name for s in destination.data_streams + destination.event_streams]
    if my_stream.name in existing_names:
        db.rollback()
        return web.Response(text="stream with the same name exists in the destination folder",
                            status=400)
    destination.event_streams.append(my_stream)
    db.commit()
    return web.json_response({"stream": my_stream.to_json()})


async def create(request):
    db: Session = request.app["db"]
    if request.content_type != 'application/json':
        return web.Response(text='content-type must be application/json', status=400)
    body = await request.json()

    if 'stream' not in body:
        return web.Response(text="provide a stream", status=400)

    # find or create the destination folder
    if 'dest_path' in body:
        try:
            destination = folder.find(body['dest_path'], db, create=True)
        except ConfigurationError as e:
            return web.Response(text="Destination error: %s" % str(e), status=400)
    elif 'dest_id' in body:
        destination = db.query(Folder).get(body["dest_id"])
    else:
        return web.Response(text="specify a destination", status=400)

    try:
        new_stream = event_stream.from_json(body['stream'])
        # clear out the id's
        new_stream.id = None
        # make sure name is unique in this destination
        existing_names = [s.name for s in destination.data_streams + destination.event_streams]
        if new_stream.name in existing_names:
            raise ConfigurationError("stream with the same name exists in the folder")
        destination.event_streams.append(new_stream)
        db.commit()
    except (TypeError, ValueError) as e:
        db.rollback()
        return web.Response(text="Invalid stream JSON: %r" % e, status=400)
    except ConfigurationError as e:
        db.rollback()
        return web.Response(text="Invalid stream specification: %s" % e, status=400)
    except KeyError as e:
        db.rollback()
        return web.Response(text="Invalid or missing stream attribute: %s" % e, status=400)
    return web.json_response(data=new_stream.to_json())


async def update(request: web.Request):
    db: Session = request.app["db"]
    if request.content_type != 'application/json':
        return web.Response(text='content-type must be application/json', status=400)

    body = await request.json()
    if 'id' not in body:
        return web.Response(text="Invalid request: specify id", status=400)

    my_stream: EventStream = db.query(EventStream).get(body['id'])
    if my_stream is None:
        return web.Response(text="stream does not exist", status=404)
    if 'stream' not in body:
        return web.Response(text="Invalid request: specify stream as JSON", status=400)
    try:
        attrs = dict(body['stream'])
    except ValueError:
        return web.Response(text="error: [stream] attribute must be JSON", status=400)
    try:
        my_stream.update_attributes(attrs)
        # make sure name is unique in this destination
        existing_names = [s.name for s in
                          my_stream.folder.data_streams + my_stream.folder.event_streams
                          if s.id != my_stream.id]
        if my_stream.name in existing_names:
            raise ConfigurationError("stream with the same name exists in the folder")
        db.commit()
    except (ValueError, ConfigurationError) as e:
        db.rollback()
        return web.Response(text="Invalid stream specification: %s" % e, status=400)
    return web.json_response(data=my_stream.to_json())


async def delete(request):
    db: Session = request.app["db"]
    data_store: EventStore = request.app["event-store"]
    # find the requested stream
    if 'path' in request.query:
        my_stream = folder.find_stream_by_path(request.query['path'], db,
                                               stream_type=EventStream)
    elif 'id' in request.query:
        my_stream = db.query(EventStream).get(request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if my_stream is None:
        return web.Response(text="stream does not exist", status=404)
    await data_store.destroy(my_stream)
    db.delete(my_stream)
    db.commit()
    return web.Response(text="ok")


# ----- data actions ----

async def write_events(request):
    db: Session = request.app["db"]
    event_store: EventStore = request.app["event-store"]
    body = await request.json()

    # find the requested stream
    if 'path' in body:
        my_stream = folder.find_stream_by_path(body['path'], db,
                                               stream_type=EventStream)
    elif 'id' in body:
        my_stream = db.query(EventStream).get(body["id"])
    else:
        return web.Response(text="specify an id or a path!!", status=400)
    if my_stream is None:
        return web.Response(text="stream does not exist", status=404)
    if 'events' not in body:
        return web.Response(text="specify events to add", status=400)
    events = await event_store.upsert(my_stream, body['events'])
    return web.json_response(data={'count': len(events), 'events': events})


async def read_events(request):
    db: Session = request.app["db"]
    event_store: EventStore = request.app["event-store"]
    # find the requested stream
    if 'path' in request.query:
        my_stream = folder.find_stream_by_path(request.query['path'], db,
                                               stream_type=EventStream)
    elif 'id' in request.query:
        my_stream = db.query(EventStream).get(request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if my_stream is None:
        return web.Response(text="stream does not exist", status=404)

    # parse optional parameters
    params = {'start': None, 'end': None, 'limit': None}
    param = ""  # to appease type checker
    try:
        for param in params:
            if param in request.query:
                params[param] = int(request.query[param])
    except ValueError:
        return web.Response(text="parameter [%s] must be an int" % param, status=400)

    # make sure parameters make sense
    if ((params['start'] is not None and params['end'] is not None) and
            (params['start'] >= params['end'])):
        return web.Response(text="[start] must be < [end]", status=400)

    # handle json filter parameter
    json_filter = None
    if 'filter' in request.query and request.query['filter'] is not None and len(request.query['filter']) > 0:
        try:
            json_filter = json.loads(request.query['filter'])
            # TODO verify syntax
        except (json.decoder.JSONDecodeError, ValueError):
            return web.Response(text="invalid filter parameter", status=400)

    # handle limit parameter, default is HARD, do not return unless count < limit
    if params['limit'] is not None and 'return-subset' not in request.query:
        if params['limit'] <= 0:
            return web.Response(text="[limit] must be > 0", status=400)
        event_count = await event_store.count(my_stream, params['start'], params['end'], json_filter=json_filter)
        if event_count > params['limit']:
            # too many events, just send the count parameter
            return web.json_response(data={'count': event_count, 'events': []})
    # if return-subset, limit is SOFT, return just that many events
    limit = None
    if params['limit'] is not None and 'return-subset' in request.query:
        if params['limit'] <= 0:
            return web.Response(text="[limit] must be > 0", status=400)
        limit = params['limit']

    events = await event_store.extract(my_stream, params['start'], params['end'], limit=limit, json_filter=json_filter)
    return web.json_response(data={'count': len(events), 'events': events})


async def remove_events(request):
    db: Session = request.app["db"]
    event_store: EventStore = request.app["event-store"]
    # find the requested stream
    if 'path' in request.query:
        my_stream = folder.find_stream_by_path(request.query['path'], db,
                                               stream_type=EventStream)
    elif 'id' in request.query:
        my_stream = db.query(EventStream).get(request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if my_stream is None:
        return web.Response(text="stream does not exist", status=404)

    # parse optional parameters
    params = {'start': None, 'end': None}
    param = ""  # to appease type checker
    try:
        for param in params:
            if param in request.query:
                params[param] = int(request.query[param])
    except ValueError:
        return web.Response(text="parameter [%s] must be an int" % param, status=400)

    # make sure parameters make sense
    if ((params['start'] is not None and params['end'] is not None) and
            (params['start'] >= params['end'])):
        return web.Response(text="[start] must be < [end]", status=400)

    # handle json filter parameter
    json_filter = None
    if 'filter' in request.query and request.query['filter'] is not None and len(request.query['filter']) > 0:
        try:
            json_filter = json.loads(request.query['filter'])
            # TODO verify syntax
        except (json.decoder.JSONDecodeError, ValueError):
            return web.Response(text="invalid filter parameter", status=400)

    await event_store.remove(my_stream, params['start'], params['end'], json_filter=json_filter)
    return web.Response(text="ok")



import numpy as np
import asyncio

from . import filter_module
from joule.models.pipes import Pipe


class FIRFilterModule(filter_module.FilterModule):
    """Apply Type I or IV FIR filter to the input"""

    def make_filter(self, parsed_args) -> np.ndarray:
        # return [taps...]
        #  implement in child
        assert False, "implement in child class"  # pragma: no cover

    async def run(self, parsed_args, inputs, outputs):
        stream_in: Pipe = inputs["input"]
        stream_out: Pipe = outputs["output"]
        taps = self.make_filter(parsed_args)
        N = len(taps)
        assert N % 2 != 0, "Tap count must be odd"
        while not self.stop_requested:
            sarray_in = await stream_in.read()
            if len(sarray_in) < (N * 2):
                # check if the pipe is closed
                if stream_in.closed:  # pragma: no cover
                    return
                # check if this is the end of an interval
                # if so we can't use this data so discard it
                if stream_in.end_of_interval:
                    stream_in.consume(len(sarray_in))
                    await stream_out.close_interval()
                # not enough data, wait for more
                await asyncio.sleep(0.1)
                continue
            # allocate output array
            output_len = len(sarray_in) - N + 1
            sarray_out = np.zeros(output_len, dtype=stream_out.dtype)
            # moving average using convolution
            bound = int(N / 2)
            sarray_out['timestamp'] = sarray_in['timestamp'][bound:-bound]
            data_in = sarray_in['data']
            # run filter by column
            data_out = np.apply_along_axis(np.convolve, 0,
                                           data_in, taps,
                                           mode='valid')
            sarray_out['data'] = data_out
            await stream_out.write(sarray_out)
            # hard to isolate in test, usually hits line 30
            if stream_in.end_of_interval:  # pragma: no cover
                await stream_out.close_interval()
                # dump all of the data because we don't
                # want to mix averages across intervals
                stream_in.consume(len(sarray_in))
            else:
                stream_in.consume(output_len)



from sqlalchemy import Column, Integer, String

from joule.models.meta import Base


class Follower(Base):
    """
    Attributes:
        name (str): follower name
        key (str): API key for accessing the follower
        location (str): URL of the follower
    """
    __tablename__ = 'follower'
    __table_args__ = {"schema": "metadata"}

    id: int = Column(Integer, primary_key=True)
    name: str = Column(String, nullable=False, unique=True)
    key: str = Column(String, nullable=False, unique=True)
    location: str = Column(String, nullable=False, unique=True)

    def __repr__(self):
        return "<Follower(id=%r name=%r, location=%r, key=%r)>" % (
            self.id, self.name, self.location, self.key)

    def to_json(self):
        """
        Returns: Dictionary of Follower attributes
        """
        return {
            'id': self.id,
            'name': self.name,
            'location': self.location,
            'key': self.key
        }



class NodeInfo:

    def __init__(self, version: str, name: str, path: str, size_reserved: int,
                 size_other: int, size_free: int, size_db: int):
        self.version = version
        self.name = name
        self.path = path
        self.size_free = size_free
        self.size_reserved = size_reserved
        self.size_db = size_db
        self.size_other = size_other



import asyncio
import aiohttp
import numpy as np
import random
import time
from typing import List, Callable, Optional
import logging
from icecream import ic

from joule.models import DataStream, pipes
from joule.models.data_store import errors
from joule.models.data_store.nilmdb_helpers import compute_path, ERRORS, check_for_error

Loop = asyncio.AbstractEventLoop
log = logging.getLogger('joule')


class Inserter:

    def __init__(self, server: str, stream: DataStream, insert_period: float, cleanup_period: float,
                 session_factory: Callable[[], aiohttp.ClientSession], retry_interval=0.5):
        self.insert_url = "{server}/stream/insert".format(server=server)
        self.remove_url = "{server}/stream/remove".format(server=server)
        self.create_url = "{server}/stream/create".format(server=server)

        self.server = server  # save for initializing decimators
        self.stream = stream
        self.path = compute_path(stream)
        # add offsets to the period to distribute traffic
        self.insert_period = insert_period + insert_period * random.random() * 0.5
        self.cleanup_period = cleanup_period + cleanup_period * random.random() * 0.25
        self._get_client = session_factory
        self.decimator = None
        self.retry_interval = retry_interval

    async def run(self, pipe: pipes.Pipe) -> None:
        """insert stream data from the queue until the queue is empty"""
        # create the database path
        # lazy stream creation,
        try:
            await self._create_path()
        except asyncio.CancelledError:
            return

        cleaner_task: Optional[asyncio.Task] = None
        if self.stream.keep_us != DataStream.KEEP_ALL:
            cleaner_task = asyncio.create_task(self._clean())
            cleaner_task.set_name("NilmDB Clean Task for [%s]" % self.path)
        while True:
            try:
                async with self._get_client() as session:
                    last_ts = None
                    while True:
                        await asyncio.sleep(self.insert_period)
                        data = await pipe.read()
                        # there might be an interval break and no new data
                        if len(data) > 0:
                            if last_ts is not None:
                                start = last_ts
                            else:
                                start = data['timestamp'][0]
                            end = data['timestamp'][-1] + 1
                            last_ts = end
                            # lazy initialization of decimator
                            if self.stream.decimate and self.decimator is None:
                                self.decimator = NilmdbDecimator(self.server, self.stream, 1, 4,
                                                                 self._get_client)
                            # send the data
                            params = {"start": "%d" % start,
                                      "end": "%d" % end,
                                      "path": self.path,
                                      "binary": '1'}
                            async with session.put(self.insert_url, params=params,
                                                   data=data.tobytes()) as resp:
                                if resp.status != 200:
                                    error = await resp.text()
                                    if cleaner_task is not None:
                                        cleaner_task.cancel()
                                        await cleaner_task
                                    raise errors.DataError("NilmDB error: %s" % error)
                            # this was successful so consume the data
                            pipe.consume(len(data))
                            # decimate the data
                            if self.decimator is not None:
                                await self.decimator.process(data)
                        # check for interval breaks
                        if pipe.end_of_interval:
                            last_ts = None
                            if self.decimator is not None:
                                self.decimator.close_interval()
            except aiohttp.ClientError as e:  # pragma: no cover
                log.warning("NilmDB raw inserter error: %r, retrying request" % e)
                await asyncio.sleep(self.retry_interval)  # retry the request
            except (pipes.EmptyPipe, asyncio.CancelledError) as e:
                break  # terminate the inserter
        if cleaner_task is not None:
            cleaner_task.cancel()
            await cleaner_task

    async def _create_path(self):
        data = {"path": compute_path(self.stream),
                "layout": self.stream.layout}
        while True:
            try:
                async with self._get_client() as session:
                    async with session.post(self.create_url, data=data) as resp:
                        await check_for_error(resp, ignore=[ERRORS.STREAM_ALREADY_EXISTS])
                        break
            except aiohttp.ClientError as e:  # pragma: no cover
                log.warning("NilmDB inserter create_path error: %r, retrying request" % e)
                await asyncio.sleep(self.retry_interval)  # retry the request

    async def _clean(self):
        while True:
            try:
                async with self._get_client() as session:
                    while True:
                        await asyncio.sleep(self.cleanup_period)
                        keep_time = int(time.time() * 1e6) - self.stream.keep_us
                        # remove raw data
                        params = {"start": "%d" % 0,
                                  "end": "%d" % keep_time,
                                  "path": self.path}
                        async with session.post(self.remove_url, params=params) as resp:
                            if resp.status != 200:  # pragma: no cover
                                msg = await resp.text()
                                log.error("NilmDB cleaning error: %s" % msg)
                        # remove decimation data
                        if self.decimator is not None:
                            for path in self.decimator.get_paths():
                                params["path"] = path
                                async with session.post(self.remove_url, params=params) as resp:
                                    if resp.status != 200:  # pragma: no cover
                                        msg = await resp.text()
                                        log.error("NilmDB cleaning error: %s" % msg)
            except aiohttp.ClientError as e:  # pragma: no cover
                log.warning("NilmDB cleaning error: %r" % e)
            except asyncio.CancelledError:
                break


class NilmdbDecimator:

    def __init__(self, server: str, stream: DataStream, from_level: int, factor: int,
                 session_factory: Callable[[], aiohttp.ClientSession], retry_interval=0.5):
        self.stream = stream
        self.level = from_level * factor
        self.insert_url = "{server}/stream/insert".format(server=server)
        self.create_url = "{server}/stream/create".format(server=server)
        self.server = server
        self.path = compute_path(stream, self.level)
        if from_level > 1:
            self.again = True
        else:
            self.again = False
        self.factor = factor
        self.layout = stream.decimated_layout
        self.buffer = []
        self.last_ts = None
        self.path_created = False
        self.child: NilmdbDecimator = None
        self.retry_interval = retry_interval
        self._get_client = session_factory
        # hold off to rate limit NilmDB traffic
        self.holdoff = 0  # random.random()

    async def process(self, data: np.ndarray) -> None:
        """decimate data and insert it, retry on error"""
        while True:
            try:
                if not self.path_created:
                    await self._create_path()
                    self.path_created = True

                async with self._get_client() as session:
                    decim_data = self._process(data)
                    if len(decim_data) == 0:
                        return
                    if self.last_ts is not None:
                        start = self.last_ts
                    else:
                        start = decim_data['timestamp'][0]
                    end = decim_data['timestamp'][-1] + 1
                    self.last_ts = end
                    # lazy initialization of child
                    if self.child is None:
                        self.child = NilmdbDecimator(self.server, self.stream, self.level,
                                                     self.factor, self._get_client)
                    params = {"start": "%d" % start,
                              "end": "%d" % end,
                              "path": self.path,
                              "binary": '1'}
                    async with session.put(self.insert_url, params=params,
                                           data=decim_data.tobytes()) as resp:
                        if resp.status != 200:  # pragma: no cover
                            error = await resp.text()
                            raise errors.DataError("NilmDB(d) error: %s" % error)
                    # feed data to child decimator
                    await self.child.process(decim_data)
                    await asyncio.sleep(self.holdoff)
                    break  # success, leave the loop
            except aiohttp.ClientError as e:  # pragma: no cover
                log.warning("NilmDB decimation error: %r, retrying request" % e)
                await asyncio.sleep(self.retry_interval)  # retry the request
            except asyncio.CancelledError:  # pragma: no cover
                break

    def close_interval(self):
        self.buffer = []
        self.last_ts = None
        if self.child is not None:
            self.child.close_interval()

    def get_paths(self) -> List[str]:
        paths = [compute_path(self.stream, self.level)]
        if self.child is not None:
            paths = paths + self.child.get_paths()
        return paths

    def _process(self, sarray: np.ndarray) -> np.ndarray:

        # flatten structured array
        data = np.c_[sarray['timestamp'][:, None], sarray['data']]

        # check if there is old data
        if len(self.buffer) != 0:
            # append the new data onto the old data
            data = np.concatenate((self.buffer, data))
        (n, m) = data.shape

        # Figure out which columns to use as the input for mean, min, and max,
        # depending on whether this is the first decimation or we're decimating
        # again.  Note that we include the timestamp in the means.
        if self.again:
            c = (m - 1) // 3
            # e.g. c = 3
            # ts mean1 mean2 mean3 min1 min2 min3 max1 max2 max3
            mean_col = slice(0, c + 1)
            min_col = slice(c + 1, 2 * c + 1)
            max_col = slice(2 * c + 1, 3 * c + 1)
        else:
            mean_col = slice(0, m)
            min_col = slice(1, m)
            max_col = slice(1, m)

        # Discard extra rows that aren't a multiple of factor
        n = n // self.factor * self.factor

        if n == 0:  # not enough data to work with, save it for later
            self.buffer = data
            return np.array([])

        trunc_data = data[:n, :]
        # keep the leftover data
        self.buffer = np.copy(data[n:, :])

        # Reshape it into 3D so we can process 'factor' rows at a time
        trunc_data = trunc_data.reshape(n // self.factor, self.factor, m)

        # Fill the result
        out = np.c_[np.mean(trunc_data[:, :, mean_col], axis=1),
                    np.min(trunc_data[:, :, min_col], axis=1),
                    np.max(trunc_data[:, :, max_col], axis=1)]

        # structure the array
        width = np.shape(out)[1] - 1
        dtype = np.dtype([('timestamp', '<i8'), ('data', '<f4', width)])
        sout = np.zeros(out.shape[0], dtype=dtype)
        sout['timestamp'] = out[:, 0]
        sout['data'] = out[:, 1:]
        # insert the data into the database
        return sout

    async def _create_path(self):
        data = {"path": compute_path(self.stream, self.level),
                "layout": self.stream.decimated_layout}
        while True:
            try:
                async with self._get_client() as session:
                    async with session.post(self.create_url, data=data) as resp:
                        await check_for_error(resp, ignore=[ERRORS.STREAM_ALREADY_EXISTS])
                        break
            except aiohttp.ClientError as e:  # pragma: no cover
                log.warning("NilmDB decimator create_path error: %r, retrying request" % e)
                await asyncio.sleep(self.retry_interval)  # retry the request



import click
import datetime
import asyncio

from joule import errors
from joule.cli.config import pass_config


@click.command(name="info")
@click.argument("path")
@pass_config
def cli_info(config, path):
    """Display event stream information."""
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            _run(config.node, path))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _run(node, path):
    my_stream = await node.event_stream_get(path)
    my_info = await node.event_stream_info(my_stream)
    # display stream information
    click.echo()
    click.echo("Event Stream Information:")
    click.echo("\tName:         %s" % my_stream.name)
    click.echo("\tStart:        %s" % _display_time(my_info.start))
    click.echo("\tEnd:          %s" % _display_time(my_info.end))
    click.echo("\tRows:         %d" % my_info.event_count)


def _display_time(time: int) -> str:
    if time is None:
        return u"\u2014"  # emdash
    return str(datetime.datetime.fromtimestamp(time / 1e6))


def _optional_field(value: str) -> str:
    if value is None or value == "":
        return u"\u2014"  # emdash
    else:
        return value



import argparse
import sys
import os
import configparser
import dateparser
import re
from typing import Tuple, Dict

from joule.errors import ConfigurationError
from joule.models import data_stream
from joule.api import data_stream as api_stream
from joule.services.helpers import load_configs

""" helpers for handling module arguments
    Complicated because we want to look for module_config
    and add these args but this should be transparent
    (eg, the -h flag should show the real help)
"""


def module_args():
    # build a dummy parser to look for module_config
    temp_parser = argparse.ArgumentParser()
    temp_parser.add_argument("--module_config", default="unset")
    arg_list = sys.argv[1:]  # ignore the program name
    stdout = sys.stdout
    sys.stdout = open(os.devnull, 'w')
    try:
        args = temp_parser.parse_known_args()[0]
        if args.module_config != "unset":
            arg_list += _append_args(args.module_config)
    except SystemExit:
        pass
    finally:
        sys.stdout.close()
        sys.stdout = stdout
    return arg_list


def load_args_from_file(path):
    args = ["--module_config", path]
    return args + _append_args(path)


def _append_args(module_config_file):
    from joule.models import module
    # if a module_config is specified add its args
    module_config = configparser.ConfigParser()
    with open(module_config_file, 'r') as f:
        module_config.read_file(f)
    my_module = module.from_config(module_config)
    args = []
    for (arg, value) in my_module.arguments.items():
        args += ["--" + arg, value]
    return args


def read_module_config(file_path: str) -> configparser.ConfigParser:
    config = configparser.ConfigParser()
    try:
        # can only happen if process does not have read permissions
        if len(config.read(file_path)) != 1:
            raise ConfigurationError("Cannot read module config [%s]" % file_path)
    except configparser.Error as e:
        raise ConfigurationError("Invalid module config: %s" % str(e))
    return config


def read_stream_configs(file_path: str) -> Dict[str, data_stream.DataStream]:
    configs = load_configs(file_path)
    streams: Dict[str, data_stream.DataStream] = {}
    for file_path, data in configs.items():
        try:
            s = data_stream.from_config(data)
            # convert to an API object
            api_s = api_stream.from_json(s.to_json())
            stream_path = _create_path(data)

            streams[stream_path] = api_s
        except KeyError as e:
            raise ConfigurationError("Invalid stream [%s]: [Main] missing %s" %
                                     (file_path, e.args[0]))
    return streams


def validate_time_bounds(start: str, end: str) -> Tuple[int, int]:
    if start is not None:
        start = int(dateparser.parse(start).timestamp() * 1e6)
    if end is not None:
        end = int(dateparser.parse(end).timestamp() * 1e6)
    if start is not None and end is not None:
        if start >= end:
            raise ConfigurationError("start [%d] must be < end [%d]" % (start, end))
    return start, end


def _create_path(data: configparser.ConfigParser) -> str:
    #
    path = data['Main']['path']
    if path != '/' and re.fullmatch(r'^(/[\w -]+)+$', path) is None:
        raise ConfigurationError(
            "invalid path, use format: /dir/subdir/../file. "
            "valid characters: [0-9,A-Z,a-z,_,-, ]")

    return '/'.join([path, data['Main']['name']])



import heapq


# Adapted from Jim Paris

def interval_difference(src, dest):
    return _interval_math_helper(src, dest, (lambda src, dest: src and not dest))


def interval_intersection(src, dest):
    return _interval_math_helper(src, dest, (lambda src, dest: src and dest))


def _interval_math_helper(src, dest, op):
    """Helper for set_difference, intersection functions,
    to compute interval subsets based on a math operator on ranges
    present in A and B.  Subsets are computed from A, or new intervals
    are generated if subset = False."""

    result = []

    # Iterate through all starts and ends in sorted order.  Add a
    # tag to the iterator so that we can figure out which one they
    # were, after sorting.
    def decorate(intervals, key_start, key_end):
        for interval in intervals:
            yield interval[0], key_start, interval
            yield interval[1], key_end, interval

    src_iter = decorate(iter(src), 0, 2)
    dest_iter = decorate(iter(dest), 1, 3)

    # Now iterate over the timestamps of each start and end.
    # At each point, evaluate which type of end it is, to determine
    # how to build up the output intervals.
    in_src = False
    in_dest = False
    out_start = None
    for (ts, k, i) in imerge(src_iter, dest_iter):
        if k == 0:
            in_src = True
        elif k == 1:
            in_dest = True
        elif k == 2:
            in_src = False
        elif k == 3:
            in_dest = False
        include = op(in_src, in_dest)
        if include and out_start is None:
            out_start = ts
        elif not include:
            if out_start is not None and out_start != ts:
                result.append([out_start, ts])
            out_start = None

    return result


# Iterator merging, based on http://code.activestate.com/recipes/491285/
def imerge(*iterables):
    """Merge multiple sorted inputs into a single sorted output.

    Equivalent to:  sorted(itertools.chain(*iterables))

    >>> list(imerge([1,3,5,7], [0,2,4,8], [5,10,15,20], [], [25]))
    [0, 1, 2, 3, 4, 5, 5, 7, 8, 10, 15, 20, 25]

    """
    heappop, siftup, _Stop = heapq.heappop, heapq._siftup, StopIteration

    h = []
    h_append = h.append
    for it in map(iter, iterables):
        try:
            next = it.__next__
            h_append([next(), next])
        except _Stop:
            pass
    heapq.heapify(h)

    while 1:
        try:
            while 1:
                v, next = s = h[0]  # raises IndexError when h is empty
                yield v
                s[0] = next()  # raises StopIteration when exhausted
                siftup(h, 0)  # restore heap condition
        except _Stop:
            heappop(h)  # remove empty iterator
        except IndexError:
            return



from typing import Union, List, Optional, Dict
from asyncio import AbstractEventLoop
import sqlalchemy

from .node_info import NodeInfo

from joule.api.session import BaseSession, TcpSession

from joule.api.folder import (Folder,
                              folder_root,
                              folder_get,
                              folder_move,
                              folder_delete,
                              folder_update)

from joule.api.module import (Module,
                              module_get,
                              module_list,
                              module_logs)

from joule.api.data_stream import (DataStream,
                                   DataStreamInfo,
                                   data_stream_get,
                                   data_stream_update,
                                   data_stream_move,
                                   data_stream_delete,
                                   data_stream_create,
                                   data_stream_info,
                                   data_stream_annotation_delete)

from joule.api.event_stream import (EventStream,
                                    EventStreamInfo,
                                    event_stream_get,
                                    event_stream_update,
                                    event_stream_move,
                                    event_stream_delete,
                                    event_stream_create,
                                    event_stream_info,
                                    event_stream_write,
                                    event_stream_read,
                                    event_stream_remove)
from joule.api.event import Event

from joule.api.data import (data_write,
                            data_read,
                            data_subscribe,
                            data_delete,
                            data_intervals,
                            data_consolidate)

from joule.api.proxy import (proxy_list,
                             proxy_get,
                             Proxy)

from joule.api.master import (master_add,
                              master_delete,
                              master_list,
                              Master)

from joule.api.annotation import (annotation_create,
                                  annotation_delete,
                                  annotation_update,
                                  annotation_get,
                                  Annotation)

from joule.api.db import (db_connect,
                          db_connection_info)

from joule.models.pipes import Pipe
from joule.utilities import ConnectionInfo


class BaseNode:

    def __init__(self, name: str,
                 session: BaseSession):
        self.name = name
        self.session = session
        # configured in child implementations
        self.url = "<unset>"

    async def close(self):
        await self.session.close()

    async def info(self) -> NodeInfo:
        resp = await self.session.get("/version.json")
        db_info = await self.session.get("/dbinfo")
        return NodeInfo(version=resp["version"], name=resp["name"], path=db_info['path'],
                        size_other=db_info["other"], size_reserved=db_info["reserved"],
                        size_free=db_info["free"], size_db=db_info["size"])

    # Folder actions

    async def folder_root(self) -> Folder:
        return await folder_root(self.session)

    async def folder_get(self,
                         folder: Union[Folder, str, int]) -> Folder:
        return await folder_get(self.session, folder)

    async def folder_move(self,
                          source: Union[Folder, str, int],
                          destination: Union[Folder, str, int]) -> None:
        await folder_move(self.session, source, destination)

    async def folder_update(self,
                            folder: Folder) -> None:
        await folder_update(self.session, folder)

    async def folder_delete(self,
                            folder: Union[Folder, str, int],
                            recursive: bool = False) -> None:
        await folder_delete(self.session, folder, recursive)

    # DataStream actions

    async def data_stream_get(self,
                              stream: Union[DataStream, str, int]) -> DataStream:
        return await data_stream_get(self.session, stream)

    async def data_stream_move(self,
                               stream: Union[DataStream, str, int],
                               folder: Union[Folder, str, int]) -> None:
        return await data_stream_move(self.session, stream, folder)

    async def data_stream_update(self,
                                 stream: DataStream) -> None:
        return await data_stream_update(self.session,
                                        stream)

    async def data_stream_delete(self,
                                 stream: Union[DataStream, str, int]) -> None:
        await data_stream_delete(self.session, stream)

    async def data_stream_create(self,
                                 stream: DataStream,
                                 folder: Union[Folder, str, int]) -> DataStream:
        return await data_stream_create(self.session, stream, folder)

    async def data_stream_info(self,
                               stream: Union[DataStream, str, int]) -> DataStreamInfo:
        return await data_stream_info(self.session, stream)

    async def data_stream_annotation_delete(self,
                                            stream: Union[DataStream, str, int],
                                            start: Optional[int] = None,
                                            end: Optional[int] = None):
        return await data_stream_annotation_delete(self.session, stream, start, end)

    # EventStream actions

    async def event_stream_get(self,
                               stream: Union[EventStream, str, int],
                               create: bool = False,
                               description: str = "",
                               event_fields=None) -> EventStream:
        return await event_stream_get(self.session, stream, create, description, event_fields)

    async def event_stream_move(self,
                                stream: Union[DataStream, str, int],
                                folder: Union[Folder, str, int]) -> None:
        return await event_stream_move(self.session, stream, folder)

    async def event_stream_update(self,
                                  stream: EventStream) -> None:
        return await event_stream_update(self.session,
                                         stream)

    async def event_stream_delete(self,
                                  stream: Union[EventStream, str, int]) -> None:
        await event_stream_delete(self.session, stream)

    async def event_stream_create(self,
                                  stream: EventStream,
                                  folder: Union[Folder, str, int]) -> EventStream:
        return await event_stream_create(self.session, stream, folder)

    async def event_stream_info(self,
                                stream: Union[EventStream, str, int]) -> EventStreamInfo:
        return await event_stream_info(self.session, stream)

    async def event_stream_write(self,
                                 stream: Union[EventStream, str, int],
                                 events: List[Event]) -> List[Event]:
        return await event_stream_write(self.session, stream, events)

    async def event_stream_read(self,
                                stream: Union[EventStream, str, int],
                                start: Optional[int] = None,
                                end: Optional[int] = None,
                                limit: Optional[int] = None,
                                json_filter: Optional[Dict[str, str]] = None) -> List[Event]:
        return await event_stream_read(self.session, stream, start, end, limit, json_filter)

    async def event_stream_remove(self,
                                  stream: Union[EventStream, str, int],
                                  start: Optional[int] = None,
                                  end: Optional[int] = None,
                                  json_filter=None) -> None:
        return await event_stream_remove(self.session, stream, start, end, json_filter)

    # Data actions

    async def data_read(self,
                        stream: Union[DataStream, str, int],
                        start: Optional[int] = None,
                        end: Optional[int] = None,
                        max_rows: Optional[int] = None) -> Pipe:
        return await data_read(self.session, stream, start, end,
                               max_rows)

    async def data_subscribe(self,
                             stream: Union[DataStream, str, int]) -> Pipe:
        return await data_subscribe(self.session, stream)

    async def data_intervals(self,
                             stream: Union[DataStream, str, int],
                             start: Optional[int] = None,
                             end: Optional[int] = None) -> List:
        return await data_intervals(self.session, stream, start, end)

    async def data_consolidate(self,
                               stream: Union[DataStream, str, int],
                               max_gap: int,
                               start: Optional[int] = None,
                               end: Optional[int] = None) -> List:
        return await data_consolidate(self.session, stream, start, end, max_gap)

    async def data_write(self, stream: Union[DataStream, str, int],
                         start: Optional[int] = None,
                         end: Optional[int] = None) -> Pipe:
        return await data_write(self.session, stream, start, end)

    async def data_delete(self, stream: Union[DataStream, str, int],
                          start: Optional[int] = None,
                          end: Optional[int] = None) -> None:
        return await data_delete(self.session, stream, start, end)

    # Module actions

    async def module_list(self,
                          statistics: bool = False) -> List[Module]:
        return await module_list(self.session, statistics)

    async def module_get(self,
                         module: Union[Module, str, int],
                         statistics: bool = False) -> Module:
        return await module_get(self.session, module, statistics)

    async def module_logs(self,
                          module: Union[Module, str, int]) -> List[str]:
        return await module_logs(self.session, module)

    # Proxy actions

    async def proxy_list(self) -> List[Proxy]:
        return await proxy_list(self.session)

    async def proxy_get(self,
                        proxy: Union[Proxy, str, int]) -> Proxy:
        return await proxy_get(self.session, proxy)

    # Master actions

    async def master_list(self) -> List[Master]:
        return await master_list(self.session)

    async def master_delete(self, master_type: str, name: str) -> None:
        return await master_delete(self.session, master_type, name)

    async def master_add(self, master_type, identifier,
                         lumen_parameters: Optional[Dict] = None,
                         api_key=None) -> Master:
        return await master_add(self.session, master_type, identifier, lumen_parameters, api_key)

    # Follower actions
    async def follower_list(self) -> List['BaseNode']:
        pass

    async def follower_delete(self, node: Union['BaseNode', str]):
        if type(node) is not str:
            name = node.name
        else:
            name = node
        data = {"name": name}
        await self.session.delete("/follower.json", params=data)

    # Annotation actions
    async def annotation_get(self, stream: Union[int, str, DataStream],
                             start: Optional[int] = None,
                             end: Optional[int] = None) -> List[Annotation]:
        return await annotation_get(self.session, stream, start, end)

    async def annotation_create(self,
                                annotation: Annotation,
                                stream: Union[int, str, DataStream], ) -> Annotation:
        return await annotation_create(self.session,
                                       annotation, stream)

    async def annotation_delete(self,
                                annotation: Union[int, Annotation]):
        return await annotation_delete(self.session, annotation)

    async def annotation_update(self,
                                annotation: Annotation):
        return await annotation_update(self.session, annotation)

    # Database actions
    async def db_connect(self) -> sqlalchemy.engine.Engine:
        return await db_connect(self.session)

    async def db_connection_info(self) -> ConnectionInfo:
        return await db_connection_info(self.session)



import click
import asyncio

from joule import errors
from joule.api import BaseNode
from joule.cli.config import pass_config


@click.command(name="delete")
@click.argument("name")
@pass_config
def follower_delete(config, name):
    """Remove a follower"""
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(config.node.follower_delete(name))
        click.echo("Follower [%s] removed" % name)
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()



import click
import asyncio

from joule.cli.config import pass_config
from joule import errors


@click.command(name="list")
@pass_config
def cli_list(config):
    """Display all users and nodes with access to the current node."""
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            _run(config.node))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _run(node):
    click.echo("This node can be controlled by:")
    masters = await node.master_list()
    # display module information
    users = [m.name for m in masters if m.master_type == 'USER']
    joule_nodes = [m.name for m in masters if m.master_type == 'JOULE_NODE']
    lumen_nodes = [m.name for m in masters if m.master_type == 'LUMEN_NODE']
    click.echo("Users:")
    if len(users) > 0:
        for user in users:
            click.echo("\t%s" % user)
    else:
        click.echo("\t[None]")
    click.echo("Joule Nodes:")
    if len(joule_nodes) > 0:
        for node in joule_nodes:
            click.echo("\t%s" % node)
    else:
        click.echo("\t[None]")
    click.echo("Lumen Nodes:")
    if len(lumen_nodes) > 0:
        for node in lumen_nodes:
            click.echo("\t%s" % node)
    else:
        click.echo("\t[None]")



from aiohttp import web
from sqlalchemy.orm import Session
from joule.models import (
    DataStream,
    EventStream,
    DataStore,
    data_stream)

from joule.models import folder, Folder
from joule.errors import ConfigurationError


async def info(request: web.Request):
    db: Session = request.app["db"]
    data_store: DataStore = request.app["data-store"]
    if 'path' in request.query:
        my_stream = folder.find_stream_by_path(request.query['path'], db, stream_type=DataStream)
    elif 'id' in request.query:
        my_stream = db.query(DataStream).get(request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if my_stream is None:
        return web.Response(text="stream does not exist", status=404)
    stream_info = await data_store.info([my_stream])
    return web.json_response(my_stream.to_json(stream_info))


async def move(request: web.Request):
    db: Session = request.app["db"]
    if request.content_type != 'application/json':
        return web.Response(text='content-type must be application/json', status=400)
    body = await request.json()
    # find the stream
    if 'src_path' in body:
        my_stream = folder.find_stream_by_path(body['src_path'], db, stream_type=DataStream)
    elif 'src_id' in body:
        my_stream = db.query(DataStream).get(body["src_id"])
    else:
        return web.Response(text="specify a source id or a path", status=400)
    if my_stream is None:
        return web.Response(text="stream does not exist", status=404)
    if my_stream.locked:
        return web.Response(text="locked streams cannot be moved", status=400)
    # find or create the destination folder
    if 'dest_path' in body:
        try:
            destination = folder.find(body['dest_path'], db, create=True)
        except ConfigurationError as e:
            return web.Response(text="Destination error: %s" % str(e), status=400)
    elif 'dest_id' in body:
        destination = db.query(Folder).get(body["dest_id"])
    else:
        return web.Response(text="specify a destination", status=400)
    # make sure name is unique in this destination
    existing_names = [s.name for s in destination.data_streams + destination.event_streams]
    if my_stream.name in existing_names:
        db.rollback()
        return web.Response(text="stream with the same name exists in the destination folder",
                            status=400)
    destination.data_streams.append(my_stream)
    db.commit()
    return web.json_response({"stream": my_stream.to_json()})


async def create(request):
    db: Session = request.app["db"]
    if request.content_type != 'application/json':
        return web.Response(text='content-type must be application/json', status=400)
    body = await request.json()

    if 'stream' not in body:
        return web.Response(text="provide a stream", status=400)

    # find or create the destination folder
    if 'dest_path' in body:
        try:
            destination = folder.find(body['dest_path'], db, create=True)
        except ConfigurationError as e:
            return web.Response(text="Destination error: %s" % str(e), status=400)
    elif 'dest_id' in body:
        destination = db.query(Folder).get(body["dest_id"])
    else:
        return web.Response(text="specify a destination", status=400)

    try:
        new_stream = data_stream.from_json(body['stream'])
        # make sure stream has at least 1 element
        if len(new_stream.elements) == 0:
            raise ConfigurationError("stream must have at least one element")

        # clear out the status flags
        new_stream.is_configured = False
        new_stream.is_destination = False
        new_stream.is_source = False
        # make sure element names are unique
        element_names = [e.name for e in new_stream.elements]
        if len(set(element_names)) != len(element_names):
            raise ConfigurationError("element names must be unique")
        # clear out the id's
        new_stream.id = None
        for elem in new_stream.elements:
            elem.id = None
        # make sure name is unique in this destination
        existing_names = [s.name for s in destination.data_streams + destination.event_streams]
        if new_stream.name in existing_names:
            raise ConfigurationError("stream with the same name exists in the folder")
        destination.data_streams.append(new_stream)
        db.commit()
    except (TypeError, ValueError) as e:
        db.rollback()
        return web.Response(text="Invalid stream JSON: %r" % e, status=400)
    except ConfigurationError as e:
        db.rollback()
        return web.Response(text="Invalid stream specification: %s" % e, status=400)
    except KeyError as e:
        db.rollback()
        return web.Response(text="Invalid or missing stream attribute: %s" % e, status=400)
    return web.json_response(data=new_stream.to_json())


async def update(request: web.Request):
    db: Session = request.app["db"]
    if request.content_type != 'application/json':
        return web.Response(text='content-type must be application/json', status=400)

    body = await request.json()
    if 'id' not in body:
        return web.Response(text="Invalid request: specify id", status=400)

    my_stream: DataStream = db.query(DataStream).get(body['id'])
    if my_stream is None:
        return web.Response(text="stream does not exist", status=404)
    if my_stream.locked:
        return web.Response(text="stream is locked", status=400)
    if 'stream' not in body:
        return web.Response(text="Invalid request: specify stream as JSON", status=400)
    try:
        attrs = dict(body['stream'])
    except ValueError:
        return web.Response(text="error: [stream] attribute must be JSON", status=400)
    try:
        my_stream.update_attributes(attrs)
        # make sure name is unique in this destination
        existing_names = [s.name for s in
                          my_stream.folder.data_streams + my_stream.folder.event_streams
                          if s.id != my_stream.id]
        if my_stream.name in existing_names:
            raise ConfigurationError("stream with the same name exists in the folder")
        db.commit()
    except (ValueError, ConfigurationError) as e:
        db.rollback()
        return web.Response(text="Invalid stream specification: %s" % e, status=400)
    return web.json_response(data=my_stream.to_json())


async def delete(request):
    db: Session = request.app["db"]
    data_store: DataStore = request.app["data-store"]
    # find the requested stream
    if 'path' in request.query:
        my_stream = folder.find_stream_by_path(request.query['path'], db, stream_type=DataStream)
    elif 'id' in request.query:
        my_stream = db.query(DataStream).get(request.query["id"])
    else:
        return web.Response(text="specify an id or a path", status=400)
    if my_stream is None:
        return web.Response(text="stream does not exist", status=404)
    if my_stream.locked or my_stream.active:
        return web.Response(text="locked streams cannot be deleted", status=400)
    await data_store.destroy(my_stream)
    db.delete(my_stream)
    db.commit()
    return web.Response(text="ok")



from typing import List, Union
import yarl

from joule import errors
from .session import BaseSession


class Proxy:
    """
    API Proxy model. See :ref:`sec-node-proxy-actions` for details on using the API to
    query proxies.

    Parameters:
       id (int): unique numeric ID assigned by Joule server
       name (str): proxy name
       target_url (str): URL to proxy
    """

    def __init__(self, id: int, name: str,
                 target_url: str):
        self.id = id
        self.name = name
        self.target_url = target_url

    def __repr__(self):
        return "<joule.api.Proxy id=%r name=%r target_url=%r>" % (
            self.id, self.name, self.target_url)


def from_json(json: dict) -> Proxy:
    return Proxy(id=json['id'], name=json['name'], target_url=json['url'])


async def proxy_get(session: BaseSession,
                    proxy: Union[Proxy, str, int]) -> Proxy:
    params = {}
    if type(proxy) is Proxy:
        params["id"] = proxy.id
    elif type(proxy) is str:
        params["name"] = proxy
    elif type(proxy) is int:
        params["id"] = proxy
    else:
        raise errors.ApiError("Invalid proxy datatype. Must be Proxy, Name, or ID")

    resp = await session.get("/proxy.json", params)
    return from_json(resp)


async def proxy_list(session: BaseSession) -> List[Proxy]:
    resp = await session.get("/proxies.json")
    return [from_json(item) for item in resp]



from .time import (time_now,
                   timestamp_to_human,
                   seconds_to_timestamp,
                   timestamp_to_seconds,
                   human_to_timestamp,
                   timestamp_to_datetime,
                   datetime_to_timestamp)
from .misc import yesno
from .interval_tools import interval_difference
#from .stream import build_stream
from .connection_info import ConnectionInfo



from typing import Optional, List
from asyncio import AbstractEventLoop, get_event_loop

from .base_node import BaseNode
from .tcp_node import TcpNode
from joule.api.session import UnixSession


class UnixNode(BaseNode):

    def __init__(self, name: str, path: str, cafile: str = ""):
        self._path = path
        session = UnixSession(path, cafile)
        super().__init__(name, session)
        self.url = "http://joule.localhost"

    def __repr__(self):
        return "<joule.api.node.UnixNode path=\"%s\">" % self._path

    async def follower_list(self) -> List[BaseNode]:
        resp = await self.session.get("/followers.json")
        return [TcpNode(item['name'], item['location'], item['key'], self.session.cafile) for item in resp]


import click
import asyncio

import joule.api
from joule.utilities import timestamp_to_human as ts2h
from joule.utilities import human_to_timestamp as h2ts
from joule import errors
from joule.api import get_node
from joule.cli.config import Config, pass_config


@click.command(name="copy")
@click.option('-s', "--start", help="timestamp or descriptive string")
@click.option('-e', "--end", help="timestamp or descriptive string")
@click.option('-r', "--replace", help="remove existing events in destination", is_flag=True)
@click.option('-n', '--new', help="copy starts at the last timestamp of the destination", is_flag=True)
@click.option('-d', '--destination-node', help="node name or Nilmdb URL")
@click.argument("source")
@click.argument("destination")
@pass_config
def cli_copy(config: Config, start, end, replace, new, destination_node, source, destination):
    """Copy events to a different stream."""
    try:
        if destination_node is None:
            dest_node = config.node
        else:
            dest_node = get_node(destination_node)
    except errors.ApiError:
        raise click.ClickException(f"Invalid destination node [{destination_node}]")
    try:
        asyncio.run(_run(config.node, dest_node, start, end, new, replace, source, destination))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        asyncio.run(config.close_node())
        if destination_node is not None:  # different destination node
            asyncio.run(dest_node.close())
    click.echo("OK")


async def _run(source_node,  dest_node, start, end, new, replace, source, destination):
    # make sure the time bounds make sense
    if start is not None:
        try:
            start = h2ts(start)
        except ValueError:
            raise errors.ApiError("invalid start time: [%s]" % start)
    if end is not None:
        try:
            end = h2ts(end)
        except ValueError:
            raise errors.ApiError("invalid end time: [%s]" % end)
    if (start is not None) and (end is not None) and ((end - start) <= 0):
        raise click.ClickException(f"Error: start {ts2h(start)} " +
                                   f"must be before end f{ts2h(end)}")

    # create the destination stream if necessary
    name = destination.split('/')[-1]
    path = "/".join(destination.split('/')[:-1])
    try:
        event_stream = joule.api.EventStream(name=name)
        await dest_node.event_stream_create(event_stream, path)
    except joule.errors.ApiError:
        pass  # stream already exists

    if replace:
        await dest_node.event_stream_remove(destination, start, end)

    stream_info = await source_node.event_stream_info(source)
    event_count = stream_info.event_count
    num_copied_events = 0
    with click.progressbar(length=event_count) as bar:
        while True:
            events = await source_node.event_stream_read(source, start=start, end=end,
                                                         limit=1000)
            # remove the event id's so it inserts as a new event
            for event in events:
                event.id = None
            if len(events) == 0:
                break
            await dest_node.event_stream_write(destination, events)
            num_copied_events += len(events)
            bar.update(len(events))
            start = events[-1].start_time + 1
        # bring bar up to 100%
        bar.update(event_count-num_copied_events)



import click
import asyncio
from treelib import Tree
from typing import Optional

import joule.api
from joule import errors
from joule.cli.config import Config, pass_config
from joule.cli.data.copy import _run as run_data_copy
from joule.cli.event.copy import _run as run_event_copy
from joule.api import BaseNode
from joule.api.folder import (Folder, folder_root)
from joule.api.data_stream import DataStream
from joule.api.event_stream import EventStream


@click.command(name="move")
@click.argument("source")
@click.argument("destination")
@pass_config
def move(config: Config, source: str, destination: str):
    """Move a folder to a new location."""
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            config.node.folder_move(source, destination))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()
    click.echo("OK")


@click.command(name="rename")
@click.argument("folder")
@click.argument("name")
@pass_config
def rename(config: Config, folder, name):
    """Rename a folder."""
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            _run_rename(config.node, folder, name))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()
    click.echo("OK")


async def _run_rename(node: BaseNode, folder_path: str, name: str):
    folder = await node.folder_get(folder_path)
    folder.name = name
    await node.folder_update(folder)


@click.command(name="delete")
@click.option("--recursive", "-r", is_flag=True)
@click.argument("folder")
@pass_config
def delete(config, folder, recursive):
    """Delete a folder and all contents."""
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            config.node.folder_delete(folder, recursive))
        click.echo("OK")
    except errors.ApiError as e:
        raise click.ClickException(str(e))
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


@click.command(name="list")
@click.argument("path", default="/")
@click.option("--layout", "-l", is_flag=True, help="include stream layout")
@click.option("--status", "-s", is_flag=True, help="include stream status")
@click.option("--id", "-i", is_flag=True, help="show ID's")
@pass_config
def list(config, path, layout, status, id):
    """Display folder hierarchy (directory layout)."""
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            _run_list(config.node, path, layout, status, id))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _run_list(node: BaseNode, path: str, layout: bool, status: bool, showid: bool):
    tree = Tree()
    if path == "/":
        root = await node.folder_root()
        root.name = ""  # omit root name
    else:
        root = await node.folder_get(path)
    _process_folder(tree, root, None, layout, status, showid)
    click.echo("Legend: [" + click.style("Folder", bold=True) + "] [Data Stream] ["
               + click.style("Event Stream", fg='cyan') + "]")
    if status:
        click.echo("\t" + click.style("\u25CF ", fg="green") + "active  " +
                   click.style("\u25CF ", fg="cyan") + "configured")
    click.echo(tree.show())


def _process_folder(tree: Tree, folder: Folder, parent_id,
                    layout: bool, status: bool, showid: bool):
    tag = click.style(folder.name, bold=True)
    if showid:
        tag += " (%d)" % folder.id
    identifier = "f%d" % folder.id
    tree.create_node(tag, identifier, parent_id)
    for stream in folder.data_streams:
        _process_data_stream(tree, stream, identifier, layout, status, showid)
    for stream in folder.event_streams:
        _process_event_stream(tree, stream, identifier, showid)
    for child in folder.children:
        _process_folder(tree, child, identifier, layout, status, showid)


def _process_data_stream(tree: Tree, stream: DataStream, parent_id,
                         layout: bool, status: bool, showid: bool):
    tag = stream.name
    if showid:
        tag += " (%d)" % stream.id
    if layout:
        tag += " (%s)" % stream.layout
    if status:
        if stream.active:
            tag = click.style("\u25CF ", fg="green") + tag
        elif stream.locked:
            tag = click.style("\u25CF ", fg="cyan") + tag

    identifier = "s%d" % stream.id
    tree.create_node(tag, identifier, parent_id)


def _process_event_stream(tree: Tree, stream: EventStream, parent_id, showid: bool):
    tag = stream.name
    if showid:
        tag += " (%d)" % stream.id
    tag = click.style(tag, fg="cyan")
    identifier = "e%d" % stream.id
    tree.create_node(tag, identifier, parent_id)


@click.command(name="copy")
@click.argument("source")
@click.argument("destination")
@click.option('-s', "--start", help="timestamp or descriptive string")
@click.option('-e', "--end", help="timestamp or descriptive string")
@click.option('-n', '--new', help="copy starts at the last timestamp of the destination", is_flag=True)
@click.option("-d", "--destination-node")
@pass_config
def copy(config, source, destination, start, end, new, destination_node):
    """Recursively copy a folder to a new location"""
    loop = asyncio.get_event_loop()
    try:
        if destination_node is not None:
            destination_node = joule.api.get_node(destination_node)
        else:
            destination_node = config.node
        loop.run_until_complete(
            _run_copy(config.node, source, destination, start, end, new, destination_node))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _run_copy(source_node, source, destination, start, end, new, destination_node) -> None:
    if type(source) is str:
        source_folder = await source_node.folder_get(source)
    else:
        source_folder = source
    for child in source_folder.children:
        await _run_copy(source_node, child, f"{destination}/{child.name}", start, end, new, destination_node)
    for data_stream in source_folder.data_streams:
        click.echo(f"Writing Data Stream {destination}/{data_stream.name}")
        await run_data_copy(source_node, start, end, new, destination_node, data_stream,
                            f"{destination}/{data_stream.name}")
    for event_stream in source_folder.event_streams:
        click.echo(f"Writing Event Stream {destination}/{event_stream.name}")
        await run_event_copy(source_node, destination_node, start, end, new, False, event_stream,
                             f"{destination}/{event_stream.name}")


@click.group(name="folder")
def folders():
    """Manage folders."""
    pass  # pragma: no cover


folders.add_command(copy)
folders.add_command(move)
folders.add_command(delete)
folders.add_command(rename)
folders.add_command(list)



import yarl


class Proxy:

    def __init__(self, name: str, uuid: int, url: yarl.URL):
        self.name = name
        self.uuid = uuid
        self.url = url

    def __repr__(self):
        return "<joule.models.Proxy name=%r uuid=%r url=%r>" % (self.name, self.uuid, self.url)



from joule.api.helpers import *
from joule.api.node import BaseNode
from joule.api.folder_type import Folder
from joule.api.event import Event
from joule.api.event_stream import (EventStream,
                                    EventStreamInfo)
from joule.api.data_stream import (DataStream,
                                   DataStreamInfo,
                                   Element)
from joule.api.module import (Module,
                              ModuleStatistics)
from joule.api.annotation import Annotation
from joule.api.proxy import Proxy



import click
import asyncio

from joule import errors
from joule.cli.config import Config, pass_config


@click.command(name="move")
@click.argument("source")
@click.argument("destination")
@pass_config
def cli_move(config: Config, source, destination):
    """Move an event stream to a different folder."""
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(
            config.node.event_stream_move(source, destination))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()
    click.echo("OK")



from sqlalchemy.orm import relationship, backref, Session
from sqlalchemy import Column, Integer, String, ForeignKey
from typing import List, TYPE_CHECKING, Optional, Dict, Union
import logging
from joule.errors import ConfigurationError
from joule.models.data_stream import DataStream
from joule.models.event_stream import EventStream

from joule.models.meta import Base

logger = logging.getLogger('joule')

Stream = Union[EventStream, DataStream]

if TYPE_CHECKING:
    from joule.models.data_store.event_store import StreamInfo as EventStreamInfo
    from joule.models.data_store.data_store import StreamInfo as DataStreamInfo


class Folder(Base):
    __tablename__ = 'folder'
    __table_args__ = {"schema": "metadata"}

    name: str = Column(String, nullable=False)
    description: str = Column(String, nullable=True)
    id: int = Column(Integer, primary_key=True)
    children: List["Folder"] = relationship("Folder",
                                            backref=backref('parent',
                                                            remote_side=[id]))
    data_streams: List[DataStream] = relationship("DataStream", back_populates="folder")
    event_streams: List[EventStream] = relationship("EventStream", back_populates="folder")
    parent_id: int = Column(Integer, ForeignKey('metadata.folder.id'))
    if TYPE_CHECKING:  # pragma: no cover
        parent: 'Folder'

    def find_stream_by_segments(self, segments: List[str]) -> Optional[Stream]:
        if len(segments) == 1:
            for stream in self.data_streams:
                if stream.name == segments[0]:
                    return stream
            for stream in self.event_streams:
                if stream.name == segments[0]:
                    return stream
            else:
                return None
        for child in self.children:
            if child.name == segments[0]:
                return child.find_stream_by_segments(segments[1:])
        else:
            return None

    def update_attributes(self, attrs: Dict) -> None:
        if 'name' in attrs:
            self.name = validate_name(attrs['name'])
        if 'description' in attrs:
            self.description = attrs['description']

    def contains_streams(self) -> bool:
        """does this folder or any of its children have streams?"""
        if len(self.data_streams) > 0:
            return True
        if len(self.event_streams) > 0:
            return True
        for c in self.children:
            if c.contains_streams():
                return True
        return False

    @property
    def locked(self):
        """is this the root folder or does this folder or any of its children have locked streams?"""
        if self.parent is None:
            return True
        for f in self.children:
            if f.locked:
                return True
        for s in self.data_streams:
            if s.locked:
                return True
        return False

    def __repr__(self):
        if self.id is None:
            return "<Folder(id=<not assigned>, name='%s')>" % self.name
        else:
            return "<Folder(id=%d, name='%s')>" % (self.id, self.name)

    def to_json(self,
                data_stream_info: Dict[int, 'DataStreamInfo'] = None,
                event_stream_info: Dict[int, 'EventStreamInfo'] = None):
        return {
            'id': self.id,
            'name': self.name,
            'description': self.description,
            'locked': self.locked,
            'children': [c.to_json(data_stream_info,
                                   event_stream_info) for c in self.children],
            'streams': [s.to_json(data_stream_info) for s in self.data_streams],
            'event_streams': [s.to_json(event_stream_info) for s in self.event_streams],
        }


def root(db: Session) -> Folder:
    root_folder = db.query(Folder). \
        filter_by(parent=None). \
        one_or_none()
    if root_folder is None:
        root_folder = Folder(name="root")
        db.add(root_folder)
        logger.info("creating root folder")
    return root_folder


def find(path: str, db: Session, create=False, parent=None) -> Optional[Folder]:
    if len(path) == 0 or path[0] != '/':
        raise ConfigurationError("invalid path [%s]" % path)
    if parent is None:
        parent = root(db)
    if path == '/':
        return parent
    path_chunks = list(reversed(path.split('/')[1:]))
    # if len(path_chunks) == 0: # never used
    #    return parent
    # make sure there are no empty chunks like /root//bad/path
    empty_chunks = [chunk for chunk in path_chunks if chunk == ""]
    if len(empty_chunks) > 0:
        raise ConfigurationError("invalid path [%s]" % path)
    name = path_chunks.pop()
    folder: Folder = db.query(Folder).filter_by(parent=parent, name=name). \
        one_or_none()
    if folder is None:
        if not create:
            return None
        folder = Folder(name=name)
        parent.children.append(folder)
        db.add(folder)
    sub_path = '/' + '/'.join(reversed(path_chunks))
    return find(sub_path, db, create=create, parent=folder)


def find_stream_by_path(path: str, db: Session, stream_type=None) -> Optional[Stream]:
    if stream_type not in [None, EventStream, DataStream]:
        raise ValueError("type must be None, EventStream, or DataStream")
    segments = path[1:].split('/')
    stream = root(db).find_stream_by_segments(segments)
    if stream_type is None:  # accept either
        return stream
    elif type(stream) is stream_type:
        return stream
    else:
        return None


# return the file path
def get_stream_path(stream: Stream) -> Optional[str]:
    if type(stream) is DataStream and stream.is_remote:
        return "[%s] %s" % (stream.remote_node, stream.remote_path)
    if stream.folder is None:
        return None

    def _get_path(folder: Folder, path: str):
        if folder.parent is None:
            return "/" + path
        return _get_path(folder.parent, folder.name + "/" + path)

    return _get_path(stream.folder, stream.name)


def validate_name(name: str) -> str:
    if name is None or len(name) == 0:
        raise ConfigurationError("missing name")
    if '/' in name:
        raise ConfigurationError("invalid name, '\\' not allowed")
    return name



from typing import Optional, Dict
import json


class Event:
    def __init__(self,
                 start_time: int,
                 end_time: Optional[int] = None,
                 content: Optional[Dict] = None,
                 id: Optional[int] = None):
        self.id = id
        self.start_time = int(start_time)
        self.end_time = int(end_time)
        if content is None:
            self.content = {}
        else:
            self.content = content

    def to_json(self):
        return {
            'id': self.id,
            'start_time': self.start_time,
            'end_time': self.end_time,
            'content': self.content
        }

    def __repr__(self):
        content_str = json.dumps(self.content)
        if len(content_str) > 20:
            content_str = content_str[:20] + " ...}"
        return "<joule.api.Event start_time: %r, end_time: %r, content: %s>" % (
            self.start_time, self.end_time, content_str)

    def __eq__(self, other):
        if other.start_time != other.start_time:
            return False
        if other.end_time != other.end_time:
            return False
        if json.dumps(other.content) != json.dumps(other.content):
            return False
        return True


def from_json(json):
    return Event(start_time=json['start_time'],
                 end_time=json['end_time'],
                 content=json['content'],
                 id=json['id'])



import click
import asyncio
import signal

from joule.cli.config import pass_config
from joule.models.pipes import compute_dtype
from joule.api.data_stream import DataStream, Element, elem_from_json

import h5py
import json
import numpy as np
from joule import errors
from joule.utilities import timestamp_to_human
from icecream import ic
stop_requested = False

BLOCK_SIZE = 10000  # insert blocks of datta


@click.command(name="ingest")
@click.option('-f', "--file", help="read data from hd5 file")
@click.option("-s", "--stream", "stream_path", help="stream path")
@pass_config
def ingest(config, stream_path, file):
    """Add data from an HD5 file to a stream."""
    signal.signal(signal.SIGTERM, handler)
    signal.signal(signal.SIGINT, handler)

    loop = asyncio.get_event_loop()

    async def _run():
        nonlocal stream_path
        # Open the file and make sure it is the right type
        try:
            hdf_root = h5py.File(file, 'r')
            hdf_timestamp = hdf_root['timestamp']
            hdf_data = hdf_root['data']
            start = hdf_timestamp[0, 0]
            end = hdf_timestamp[-1, 0]
            # make sure the length of both datasets are  the same
            if len(hdf_data) != len(hdf_timestamp):
                raise click.ClickException("Length of [data] and [timestamp] datasets must match")
            # if a stream is not specified see if one is in the data file
            if stream_path is None:
                try:
                    stream_path = hdf_root.attrs['path']
                except KeyError:
                    raise click.ClickException("Specify a target stream with --stream")
        except OSError:
            raise click.ClickException("Data file [%s] must be hdf5 format" % file)
        except KeyError:
            raise click.ClickException("Data file must contain [data] and [timestamp] datasets")

        # get the stream object from the API
        try:
            stream_obj = await config.node.data_stream_get(stream_path)
            print("Destination stream: %s" % stream_path)

            stream_info = await config.node.data_stream_info(stream_path)
            # make sure the datatypes match
            dtype = compute_dtype(stream_obj.layout)
            if dtype[1].base != hdf_data.dtype:
                raise click.ClickException("Incompatible datatypes, stream is [%s] and data file is [%s]" % (
                    (dtype[1].base, hdf_data.dtype)
                ))
            # make sure the number of elements match
            if len(stream_obj.elements) != hdf_data.shape[1]:
                raise click.ClickException("DataStream has [%d] elements but data file has [%d] elements" % (
                    len(stream_obj.elements), hdf_data.shape[1]
                ))
            # check if there is existing data in this time period
            if stream_info.rows > 0 and (
                    start < stream_info.end and end >= stream_info.start):
                # confirm overwrite
                if not click.confirm("This will remove existing data between %s- %s" % (
                        timestamp_to_human(start),
                        timestamp_to_human(end))):
                    click.echo("Cancelled")
                    return
                await config.node.data_delete(stream_obj, start, end+1)
        except errors.ApiError as e:
            if '404' not in str(e):
                raise click.ClickException(str(e))
            # this stream doesn't exist, create it from the hdf attributes
            stream_obj = await _create_stream(stream_path, hdf_root, config.node)

        pipe = await config.node.data_write(stream_obj)

        # progress bar for writing to a file
        bar_ctx = click.progressbar(length=len(hdf_data), label='ingesting data')
        bar = bar_ctx.__enter__()
        for idx in range(0, len(hdf_data), BLOCK_SIZE):
            ts = hdf_timestamp[idx:idx + BLOCK_SIZE]
            data = hdf_data[idx:idx + BLOCK_SIZE]
            sdata = np.empty(len(ts), dtype=compute_dtype(stream_obj.layout))
            sdata['timestamp'][:, None] = ts
            sdata['data'] = data
            await pipe.write(sdata)
            bar.update(len(data))
        await pipe.close()
        bar_ctx.__exit__(None, None, None)

    try:
        loop.run_until_complete(_run())
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _create_stream(stream_path, hdf_root, node):
    # try to get the elements from the hdf attrs
    try:
        element_json = json.loads(hdf_root.attrs['element_json'])
        elements = [elem_from_json(e) for e in element_json]
    except KeyError:
        # just make default elements
        num_elements = hdf_root['data'].shape[1]
        elements = [Element("Element %d" % i) for i in range(num_elements)]

    stream_name = stream_path.split('/')[-1]
    folder = '/'.join(stream_path.split('/')[:-1])
    if folder == '':
        raise click.ClickException("Invalid stream path, must include a folder")
    new_stream = DataStream(stream_name)
    new_stream.datatype = hdf_root['data'].dtype.name
    new_stream.elements = elements
    stream_obj = await node.data_stream_create(new_stream, folder)
    print("creating [%s]" % stream_path)
    return stream_obj


def _build_default_elements(num_elements):
    return [Element("Element %d" % i) for i in range(num_elements)]


def handler(signum, frame):
    global stop_requested
    stop_requested = True



import click
import asyncio
import argparse

import joule.errors
from joule.utilities.interval_tools import interval_intersection, interval_difference
from joule import errors
from joule.cli.config import Config, pass_config
from joule.utilities import timestamp_to_human, human_to_timestamp
from joule.client.builtins.merge_filter import MergeFilter


@click.command(name="merge")
@click.option("--start", help="timestamp or descriptive string")
@click.option("--end", help="timestamp or descriptive string")
@click.option('-p', "--primary", help="primary stream", required=True)
@click.option('-s', "--secondary", multiple=True, help="secondary stream")
@click.option('-d', "--destination", help="destination stream", required=True)
@pass_config
def merge(config: Config, start, end, primary, secondary, destination):
    """Merge two or more data streams. Elements will be prefixed with source stream name or a custom prefix
       using a : such as /source/streamA would create elements streamA E1, streamA E2, etc and /source/streamA:A
       would create elements A E1, A E2, etc"""
    if start is not None:
        try:
            start = human_to_timestamp(start)
        except ValueError:
            raise click.ClickException("invalid start time: [%s]" % start)
    if end is not None:
        try:
            end = human_to_timestamp(end)
        except ValueError:
            raise click.ClickException("invalid end time: [%s]" % end)

    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(_run(start, end, destination, primary, secondary, config.node))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        loop.run_until_complete(
            config.close_node())
        loop.close()


async def _run(start, end, destination, primary, secondaries, node):
    # --- RETRIEVE SOURCE STREAMS ---
    # prints out error messages if the source streams do not exist
    try:
        primary_stream: joule.DataStream = await node.data_stream_get(primary.split(':')[0])
    except joule.errors.ApiError as e:
        raise click.ClickException(f"Primary stream {primary}: {str(e)}")
    secondary_streams = []
    for s in secondaries:
        try:
            s_stream = await node.data_stream_get(s.split(':')[0])
            secondary_streams.append(s_stream)
        except joule.errors.ApiError as e:
            raise click.ClickException(f"Secondary stream {s}: {str(e)}")

    # --- CHECK FOR DESTINATION STREAM ---
    destination_width = len(primary_stream.elements)
    for s in secondary_streams:
        destination_width += len(s.elements)
    dest_exists = True
    dest_stream = None  # this is set below, or created later
    try:
        dest_stream = await node.data_stream_get(destination)
        if len(dest_stream.elements) != destination_width:
            raise click.ClickException(f"Destination must have {destination_width} elements")
        if not dest_stream.datatype.startswith('float'):
            raise click.ClickException(f"Destination must be a float datatype")

    except joule.errors.ApiError as e:
        dest_exists = False

    click.echo(f"Primary Stream: \n\t{primary}")
    click.echo(f"Secondary Streams: \n\t{', '.join(s for s in secondaries)}")
    if dest_exists:
        click.echo(f"Destination Stream: \n\t{destination}")
    else:
        click.echo(f"Creating Destination: \n\t{destination}")

    if not click.confirm("Proceed?"):
        click.echo("Cancelled")
        return

    if not dest_exists:
        # create the destination
        dest_path = destination.split('/')[:-1]
        dest_name = destination.split('/')[-1]
        elements = []
        # check for a primary prefix
        prefix = ""
        if ':' in primary:
            prefix = primary.split(':')[-1]
        else:
            prefix = primary_stream.name
        for e in primary_stream.elements:
            e.name = prefix + ' ' + e.name
            elements.append(e)
        for i in range(len(secondaries)):
            prefix = ""
            if ':' in secondaries[i]:
                prefix = secondaries[i].split(':')[-1]
            else:
                prefix = secondary_streams[i].name
            for e in secondary_streams[i].elements:
                e.name = prefix +' '+ e.name
                elements.append(e)
        dest_stream = joule.api.DataStream(name=dest_name, elements=elements, datatype='float32')
        try:
            dest_stream = await node.data_stream_create(dest_stream, '/'.join(dest_path))
        except joule.errors.ApiError as e:
            raise click.ClickException(f"Creating destination: {str(e)}")

    # --- READY TO GO ---
    common_intervals = await node.data_intervals(primary_stream, start=start, end=end)
    for s in secondary_streams:
        s_intervals = await node.data_intervals(s, start=start, end=end)
        common_intervals = interval_intersection(common_intervals, s_intervals)
    copied_intervals = await node.data_intervals(dest_stream, start=start, end=end)
    # do not copy intervals that we already have
    pending_intervals = interval_difference(common_intervals, copied_intervals)
    # only copy intervals with at least 1 second of data- there are issues with 1 sample offsets
    # that cause merge to think there is missing data when there is not any- this is probably due to
    # the backend not having the data boundaries stored exactly correctly
    pending_intervals = [i for i in pending_intervals if (i[1]-i[0]) > 1e6]
    if len(pending_intervals) == 0:
        click.echo("Destination already has all the data, nothing to do")
    else:
        start_time = joule.utilities.timestamp_to_human(pending_intervals[0][0])
        end_time = joule.utilities.timestamp_to_human(pending_intervals[-1][0])
        click.echo(
            f"Merging from {start_time} to {end_time} ({len(pending_intervals)} intervals)")
    for interval in pending_intervals:
        start = interval[0]
        end = interval[1]
        click.echo("Processing [%s] -> [%s]" % (timestamp_to_human(start), timestamp_to_human(end)))
        merge_filter = MergeFilter()
        inputs = {'primary': await node.data_read(primary_stream, start, end)}
        for i in range(len(secondary_streams)):
            inputs[f"secondary_{i}"] = await node.data_read(secondary_streams[i], start, end)
        outputs = {'destination': await node.data_write(dest_stream, start, end)}
        args = argparse.Namespace(primary="primary")
        await merge_filter.run(args, inputs, outputs)
        await outputs['destination'].close()



import click
from .list import cli_list
from .delete import cli_delete
from .add import cli_add


@click.group(name="master")
def master():
    """Manage node access."""
    pass  # pragma: no cover


master.add_command(cli_list)
master.add_command(cli_delete)
master.add_command(cli_add)



from typing import Union

from joule import errors
from .folder_type import Folder
from .session import BaseSession
from .data_stream import from_json as data_stream_from_json
from .event_stream import from_json as event_stream_from_json


def from_json(json) -> Folder:
    data_streams = [data_stream_from_json(val) for val in json['streams']]
    event_streams = [event_stream_from_json(val) for val in json['event_streams']]

    children = [from_json(val) for val in json['children']]
    my_folder = Folder()
    my_folder.id = json['id']
    my_folder.name = json['name']
    my_folder.description = json['description']
    my_folder.children = children
    my_folder.data_streams = data_streams
    my_folder.event_streams = event_streams
    my_folder.locked = json['locked']
    return my_folder


async def folder_root(session: BaseSession) -> Folder:
    resp = await session.get("/folders.json")
    return from_json(resp)


async def folder_move(session: BaseSession,
                      source: Union[Folder, str, int],
                      destination: Union[Folder, str, int]) -> None:
    data = {}

    if type(source) is Folder:
        data["src_id"] = source.id
    elif type(source) is int:
        data["src_id"] = source
    elif type(source) is str:
        data["src_path"] = source
    else:
        raise errors.ApiError("Invalid source datatype. Must be Folder, Path, or ID")

    if type(destination) is Folder:
        data["dest_id"] = destination.id
    elif type(destination) is int:
        data["dest_id"] = destination
    elif type(destination) is str:
        data["dest_path"] = destination
    else:
        raise errors.ApiError("Invalid destination datatype. Must be Folder, Path, or ID")

    await session.put("/folder/move.json", data)


async def folder_delete(session: BaseSession,
                        folder: Union[Folder, str, int],
                        recursive: bool = False) -> None:
    _recursive = 0
    if recursive:
        _recursive = 1
    data = {"recursive": _recursive}

    if type(folder) is Folder:
        data["id"] = folder.id
    elif type(folder) is int:
        data["id"] = folder
    elif type(folder) is str:
        data["path"] = folder
    else:
        raise errors.ApiError("Invalid folder datatype. Must be Folder, Path, or ID")

    await session.delete("/folder.json", data)


async def folder_update(session: BaseSession,
                        folder: Folder) -> None:
    return await session.put("/folder.json", {"id": folder.id,
                                              "folder": folder.to_json()})


async def folder_get(session: BaseSession,
                     folder: Union[Folder, str, int]) -> Folder:
    params = {}
    if type(folder) is Folder:
        params["id"] = folder.id
    elif type(folder) is int:
        params["id"] = folder
    elif type(folder) is str:
        params["path"] = folder
    else:
        raise errors.ApiError("Invalid folder datatype. Must be Folder, Path, or ID")

    resp = await session.get("/folder.json", params)
    return from_json(resp)



import click
import asyncio
from joule.cli.config import pass_config
from joule import errors, api, utilities


@click.command(name="add")
@click.argument("name")
@click.argument("url")
@click.argument("key")
@pass_config
def node_add(config, name: str, url: str, key: str):
    """Add a new node (requires API key)

    Use the information provided by 'joule master add user' """
    loop = asyncio.get_event_loop()
    my_node = None
    try:
        if not url.startswith("http"):
            url = loop.run_until_complete(
                utilities.misc.detect_url(url, 8088))
            if url is None:
                raise click.ClickException(
                    "unable to contact [%s], node not added" % url)
        my_node = api.TcpNode(name, url, key)
        api.save_node(my_node)
        click.echo("Added [%s] to authorized nodes" % name)
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        if my_node is not None:
            loop.run_until_complete(
                my_node.close())
        loop.close()
